/* comments, suggestions, complaints, etc to dsharp@interport.net */
/* http://www.interport.net/~dsharp */
#include "colors.Inc"
#include "textures.Inc"
#include "utilmac.inc"

/*
   The file 'hknot.inc' contains arrays representing
   Aaron K. Trautwein's 'Harmonic' knot series.
   prime knots up through 7 (minimal) crossings
      3_1, 4_1, 5_1, 5_2, 6_1, 6_2, 6_3,
      7_1, 7_2, 7_3, 7_4, 7_5, 7_6, 7_7,
   and
      Granny knot, Square Knot
   Use the #declares in the file hknot.inc to define the
   knot parameterization, or use your own ([3][15][2])
   array of Fourier coefficients
   Reference:
	http://www.carthage.edu/~trautwn/
	http://www.math.uiowa.edu/~jsimon/aaron.html

*/

#include "hknot.inc"

/*
 setting up an array of coefficients
 for the 'harmonic' Fourier series.
      array[3][15][2]
 '3' for the three components x,y,z
 '15' for terms up to the 15th order, and
 '2' for the cosine and sine coefficients
 for example:
  the trefoil knot (3_1) is parametrized by:
  x(m) = 41*cos(m) - 18*sin(m) - 83*cos(2*m) - 83*sin(2*m) 
     - 11*cos(3*m) + 27*sin(3*m), 
  y(m) = 36*cos(m) + 27*sin(m) - 113*cos(2*m) + 30*sin(2*m) 
     + 11*cos(3*m) - 27*sin(3*m), 
  z(m) = 0*cos(m) + 45*sin(m) - 30*cos(2*m) + 113*sin(2*m) 
     - 11*cos(3*m) + 27*sin(3*m) 

  The array for this would look like:
  #declare three_1_array=array[3][15][2]{
    //  the Fourier series coefficients of:
    //       cos(m) sin(m)  cos(2m) sin(2m)  cos(3m) sin(3m)  ...
    /* x */{ { 41,   -18  } {-83,    -83   } {-11,     27   } {0,0}{0,0}{0,0}{0,0}{0,0}{0,0}{0,0}{0,0}{0,0}{0,0}{0,0}{0,0}},
    /* y */{ { 36,    27  }  {-113,   30   } { 11,    -27   } {0,0}{0,0}{0,0}{0,0}{0,0}{0,0}{0,0}{0,0}{0,0}{0,0}{0,0}{0,0}},
    /* z */{ {  0,    45  }  {-30,   113   } {-11,     27   } 	{0,0}{0,0}{0,0}{0,0}{0,0}{0,0}{0,0}{0,0}{0,0}{0,0}{0,0}{0,0}}
  } 
  */

/*---- Random coefficients  ----------------------*/
//  This constructs a random array of coefficients.
//  Probably not very useful, but often interesting
#macro randomcoefficients(maxorder,myseed)
    #declare randarray=array[3][15][2]
    #local S1 = seed(myseed);
    #local i=0;
    #local j=0;
    #local k=0;
    #while(i<3)
	  #while(j<15)
     	#while(k<2)
               #if(j<maxorder)// put a limit on the order of our series
                  #declare randarray[i][j][k]=50-int(100*rand(S1));
               #else
                  #declare randarray[i][j][k]=0;
               #end
			#declare k=k+1;
          #end
          #declare k=0;
		#declare j=j+1;
	  #end
       #declare j=0;
       #declare i=i+1;
    #end
    randarray 
#end
/*----------------------------------*/

/* H_Array is the array of Fourier coefficients
    which we are actually going to use */

#declare H_Array=GrannyKnot_array
  
// the macro which returns coordinates of points
// on the knot's curve. 0 < m < 2*pi 
#macro knot(m)
    #local i=0;
    #local xx=0; 
    #while (i<15)
      #local xx=xx+H_Array[0][i][0]*cos((i+1)*m)+H_Array[0][i][1]*sin((i+1)*m);
      #local i=i+1; 
    #end
    #local i=0;
    #local yy=0; 
    #while (i<15)
      #local yy=yy+H_Array[1][i][0]*cos((i+1)*m)+H_Array[1][i][1]*sin((i+1)*m);
      #local i=i+1;
    #end
    #local i=0;
    #local zz=0; 
    #while (i<15)
      #local zz=zz+H_Array[2][i][0]*cos((i+1)*m)+H_Array[2][i][1]*sin((i+1)*m);
      #local i=i+1;
    #end
    <xx,yy,zz>
#end

// there are many uses for the curve's tangent vector
// returns <d(knot(m))/dx,d(knot(m))/dy,d(knot(m))/dz>
#macro d1knot(m) 
    #local i=0;
    #local xx=0; 
    #while (i<15)
      #local xx=xx-H_Array[0][i][0]*(i+1)*sin((i+1)*m)+H_Array[0][i][1]*(i+1)*cos((i+1)*m);
      #local i=i+1; 
    #end
    #local i=0;
    #local yy=0; 
    #while (i<15)
      #local yy=yy-H_Array[1][i][0]*(i+1)*sin((i+1)*m)+H_Array[1][i][1]*(i+1)*cos((i+1)*m);
      #local i=i+1;
    #end
    #local i=0;
    #local zz=0; 
    #while (i<15)
      #local zz=zz-H_Array[2][i][0]*(i+1)*sin((i+1)*m)+H_Array[2][i][1]*(i+1)*cos((i+1)*m);
      #local i=i+1;
    #end
    <xx,yy,zz>
#end 
 
// the curve's 2nd derivative (acceleration)
// for determining how the curve is curving
// returns <d2(knot(m))/dx2,d2(knot(m))/dy2,d2(knot(m))/dz2>
#macro d2knot(m) 
    #local i=0;
    #local xx=0; 
    #while (i<15)
      #local xx=xx-H_Array[0][i][0]*(i+1)*(i+1)*cos((i+1)*m)-H_Array[0][i][1]*(i+1)*(i+1)*sin((i+1)*m);
      #local i=i+1; 
    #end
    #local i=0;
    #local yy=0; 
    #while (i<15)
      #local yy=yy-H_Array[1][i][0]*(i+1)*(i+1)*cos((i+1)*m)-H_Array[1][i][1]*(i+1)*(i+1)*sin((i+1)*m);
      #local i=i+1;
    #end
    #local i=0;
    #local zz=0; 
    #while (i<15)
      #local zz=zz-H_Array[2][i][0]*(i+1)*(i+1)*cos((i+1)*m)-H_Array[2][i][1]*(i+1)*(i+1)*sin((i+1)*m);
      #local i=i+1;
    #end
    <xx,yy,zz>
#end 

/*-------------------------------*/
/*-------------------------------*/


#macro mytexture(beta)
   texture{pigment{hcolor(beta)} finish{reflection .2}}
#end

/* 
  theradius(beta) gives a smoothly varying radius or
   width for spheres, tubes, ribbons, etc
   keep RadiusFreq an integer so that the radius 
   varies smoothly at beta=0->2pi 
*/
#declare minRadius =10;
#declare maxRadius =20;
#declare RadiusPhase=pi/2;
#declare RadiusFreq=5;
#declare midRadius=(maxRadius-minRadius)/2;
#macro theradius(beta)   
    midRadius*cos(RadiusFreq*beta+RadiusPhase)+midRadius+minRadius
#end

/*-----------------------------------------------*/
/* various objects to build the knot out of. */
/* 
   mostly these macros take vectors v1 and v2
   as calling parameters.
   v1 and v2 generally refer to consecutive
   points on the knot curve. v1u, and v2u are
   (normalized) directions perpendicular to the knot
   tangent and its curvature 'radius' at v1 and v2.
   uLen is the 'width' of the curve (for tubes, ribbons, etc)
*/
// just a sphere at v1
#macro sph_object(v1,R)
// eg:  sph_object(knot(beta),sphereRadius))
	sphere{<0,0,0>,R
          translate v1
     }
#end

// a sphere at v1, and a cylinder from v1 to v2
#macro cyl_sph_object(v1,v2,R)
// eg: cyl_sph_object(knot(beta1),knot(beta2), tubeRadius)
   union{
     sphere{v1, R}
     #if( (v1.x != v2.x) | (v1.y != v2.y) | (v1.z != v2.z))
        cylinder{v1,v2,R}
     #else
        sphere{v1,R/2} //suppress csg warning
     #end
	
   }
#end

// a sphere at v1 and a cone from v1 to v2
#macro cone_sph_object(v1,R1,v2,R2)
   union{
     sphere{v1, R1}
     #if( (v1.x != v2.x) | (v1.y != v2.y) | (v1.z != v2.z))
        cone{v1,R1,v2,R2}
     #else
        sphere{v1,R1/2} //suppress 'single object' csg warning
     #end
	
   }
#end
#macro merge_cone_sph_object(v1,R1,v2,R2)
   merge{
     sphere{v1, R1}
     #if( (v1.x != v2.x) | (v1.y != v2.y) | (v1.z != v2.z))
        cone{v1,R1,v2,R2}
     #else
        sphere{v1,R1/2} //suppress 'single object' csg warning
     #end
	
   }
#end

// a pair of tracks
#macro cylribs(v1,v1u,v2,v2u,uLen,R)
   #if(!vzero_p((v1-v1u)-(v2-v2u)))
      cylinder{v1-uLen*v1u,v2-uLen*v2u,R}
      cylinder{v1+uLen*v1u,v2+uLen*v2u,R}      
   #end
   #if(!vzero_p(v1-v2))
      //cylinder{v1,v2,R}
   #end

#end

// these are four 'pigment, finish' pairs for the "array of cylinders"
#macro T1(m,P)
     pigment{hcolor(m) } finish{reflection .2}
#end
#macro T2(m,P)
    pigment{hcolor(2*m) } finish{reflection .2}
#end
#macro T3(m,P)
     pigment{hcolor(m+pi)} finish{reflection .2}
#end
#macro T4(m,P)
     pigment{hcolor((2*m+pi)) } finish{reflection .2}
#end

#macro cylarray(v1,v1u,v2,v2u,uLen,R,m)
   
   #if(!vzero_p((v1-v1u)-(v2-v2u)))
     union{
      sphere{v1-uLen*v1u,R texture{T1(m,0)}}      
      cylinder{v1-uLen*v1u,v2-uLen*v2u,R texture{T1(m,0)}}
      sphere{v1+uLen*v1u,R texture{T2(m,0)}}
      cylinder{v1+uLen*v1u,v2+uLen*v2u,R texture{T2(m,0)}}

      #local v1Left=vnormalize(vcross(v2-v1,v1u));
      #local v2Left=vnormalize(vcross(v2-v1,v2u));
      sphere{v1-uLen*v1Left,R texture{T3(m,0)}}
      cylinder{v1-uLen*v1Left,v2-uLen*v2Left,R texture{T3(m,0)}}
      sphere{v1+uLen*v1Left,R texture{T4(m,0)}}
      cylinder{v1+uLen*v1Left,v2+uLen*v2Left,R texture{T4(m,0)}}
     }
   #end

#end

#macro ktriangles(v1,v1u,v2,v2u,uLen,m)
     #local tv1=v1-uLen*v1u;
     #local tv2=v2-uLen*v2u;
     #local tv3=v1+uLen*v1u;
     #local tv4=v2+uLen*v2u;
     triangle{tv1,tv2,tv3}
     triangle{tv3,tv2,tv4}     
#end

#macro striangles(v1,v1u,v1l,v2,v2u,v2l,uLen,m)
       #local tempv1l=v1l;
       #local tempv2l=v2l;
       #local tv1=v1-uLen*v1u;
       #local tv2=v2-uLen*v2u;
       #local tv3=v1+uLen*v1u;
       #local tv4=v2+uLen*v2u;
       #if(degenerate_norms_p(tv1,tv2,tv3,v1l,v2l,v1l))
          #debug "'degenerate' normals\n"
          #declare tempv2l=v1l;
       #end
       smooth_triangle{
          tv1,v1l,
          tv2,tempv2l,
          tv3,v1l
        }
        #if(degenerate_norms_p(tv3,tv2,tv4,v1l,v2l,v2l))
          #debug "degenerate normals\n"
          #declare tempv1l=v2l; 
        #end
        smooth_triangle{
           tv3,tempv1l,
           tv2,v2l,
           tv4,v2l
        }

#end
#macro striangles2(v1,v1u,v1l,v2,v2u,v2l,uLen,m)
       #local tempv1l=v1l;
       #local tempv2l=v2l;
       #local tv1=v1-uLen*v1u;
       #local tv2=v2-uLen*v2u;
       #local tv3=v1+uLen*v1u;
       #local tv4=v2+uLen*v2u;
       #if(degenerate_norms_p(tv1,tv2,tv3,v1l,v2l,v1l))
          #debug "'degenerate' normals\n"
          triangle{tv1,tv2,tv3}
       #else
          smooth_triangle{
             tv1,v1l,
             tv2,tempv2l,
             tv3,v1l
          }
        #end
        #if(degenerate_norms_p(tv3,tv2,tv4,v1l,v2l,v2l))
          #debug "'degenerate' normals\n"
          triangle{tv3,tv2,tv4}
        #else
          smooth_triangle{
             tv3,tempv1l,
             tv2,v2l,
             tv4,v2l
          }
        #end
#end

#macro triangleribs(v1,v1u,v1l,v2,v2u,v2l,uLen,m,tradius) 
     sphere{v1+uLen*v1u,tradius}
     sphere{v1-uLen*v1u, tradius}
     sphere{v2-uLen*v2u, tradius}
     sphere{v2+uLen*v2u, tradius}
     cylinder{v1+uLen*v1u,v1-uLen*v1u, tradius}
     cylinder{v1+uLen*v1u,v2-uLen*v2u, tradius}  
     cylinder{v1-uLen*v1u,v2-uLen*v2u, tradius}
     cylinder{v1+uLen*v1u,v2+uLen*v2u, tradius}     
#end

#include "metals.inc"
// primitive 'roller coaster' tracks
#macro tracks(v1,v1u,v1l,v2,v2u,v2l,uLen,m,tradius) 
     sphere{v1+uLen*v1u,tradius} 
     sphere{v1-uLen*v1u, tradius}
     // 'tie'
     cylinder{v1+uLen*v1u,v1-uLen*v1u, tradius/2
       texture{DMFWood6 scale 10}
     }
     // 'rails'
     union{
       cylinder{v1-uLen*v1u,v2-uLen*v2u, tradius}
       cylinder{v1+uLen*v1u,v2+uLen*v2u, tradius} 
       //texture{pigment{Silver} finish{reflection .2}}
       texture {Silver1}
     }    
#end
/*---------------------------------------------------*/


// use R0 as the size of the tube, ribbon, whatever
#declare R0=15; 

// Total_Segs is the 'resolution' of the parameterization
#declare Total_Segs= 200;

#declare Twists=0;
#declare betaphase=0;
#declare dbeta=2*pi/Total_Segs;


#declare knot_thing= object{
	union{
          // start the knot at betaphase (normally 0)
          #declare lastV=knot(betaphase);          
          #declare lastR=theradius(betaphase);
          #declare Seg_Count=1;
          
          #while( Seg_Count<=Total_Segs+1)
            #declare beta = Seg_Count*2*pi/Total_Segs+betaphase;
            #declare thisV=knot(beta); 
            #declare thisLeft=vnormalize(d2knot(beta));
		  
            // extra twist:
            #if(Twists)
              #declare thisLeft=vaxis_rotate(thisLeft,thisV-lastV,180/pi*(Twists*beta));
		  #end 

            // define 'up' (relative to the path):
            #declare thisUp=vnormalize(vcross( vnormalize(d1knot(beta)),thisLeft));
            
            #declare thisR=theradius(beta); 
            
            #if (Seg_Count>1)			
            
                #declare T=texture{pigment {hcolor(beta)} finish { phong 1 phong_size 200 reflection .15 }}
                //#declare T=texture{pigment {hcolor(beta) transmit .8} finish { phong 1 phong_size 200 reflection .15 }} 
                union{ 
                
                  //ktriangles(thisV,thisUp,lastV,lastUp,R0,beta)
                  //cylribs(thisV,thisUp,lastV,lastUp,R0/2,R0/8)
                  //cylarray(thisV,thisUp,lastV,lastUp,R0/2,R0/4,beta)
                  //striangles(lastV,lastUp,lastLeft,thisV,thisUp,thisLeft,R0,beta)
                  //striangles2(lastV,lastUp,lastLeft,thisV,thisUp,thisLeft,R0,beta)			  
                  tracks(lastV,lastUp,lastLeft,thisV,thisUp,thisLeft,R0/2,beta,3)
                  //triangleribs(lastV,lastUp,lastLeft,thisV,thisUp,thisLeft,R0,beta,3)
			   
			   texture {T}
			   	 
                }
            #end      
	                     
            #declare lastV=thisV;
            #declare lastR=thisR; 
            #declare lastUp= thisUp;
            #declare lastLeft=thisLeft; 
           
            #declare Seg_Count=Seg_Count+1;
          #end // end of the Seg_Count while loop
	}// end of union
}

// viewpoint location needs adjusting for different knots
camera{
	location <450.00, 00, 00>
	look_at <0.00, 00, 0.00>
}

light_source {
	<1100, 100, -200>
	color White
}
background{White}

//actually put the knot in the scene
object{
	knot_thing
	rotate<0,80,0>
}
